@@ -90,6 +90,9 @@
 #endif
 #endif
 #endif
+#ifndef MMAP_LIMIT
+#define MMAP_LIMIT              (4*1024*1024)
+#endif
 
 /* Server core module... This module provides support for really basic
  * server operations, including options and commands which control the
@@ -127,7 +130,7 @@ static void *create_core_dir_config(pool *a, char *dir)
 
     conf->content_md5 = 2;
 
-    conf->use_canonical_name = 1 | 2;	/* 2 = unset, default on */
+    conf->use_canonical_name = USE_CANONICAL_NAME_UNSET;
 
     conf->hostname_lookups = HOSTNAME_LOOKUP_UNSET;
     conf->do_rfc1413 = DEFAULT_RFC1413 | 2; /* set bit 1 to indicate default */
@@ -148,6 +151,12 @@ static void *create_core_dir_config(pool *a, char *dir)
 #ifdef WIN32
     conf->script_interpreter_source = INTERPRETER_SOURCE_UNSET;
 #endif
+
+    conf->server_signature = srv_sig_unset;
+
+    conf->add_default_charset = ADD_DEFAULT_CHARSET_UNSET;
+    conf->add_default_charset_name = DEFAULT_ADD_DEFAULT_CHARSET_NAME;
+
     return (void *)conf;
 }
 
@@ -236,7 +245,7 @@ static void *merge_core_dir_configs(pool *a, void *basev, void *newv)
     if ((new->content_md5 & 2) == 0) {
         conf->content_md5 = new->content_md5;
     }
-    if ((new->use_canonical_name & 2) == 0) {
+    if (new->use_canonical_name != USE_CANONICAL_NAME_UNSET) {
 	conf->use_canonical_name = new->use_canonical_name;
     }
 
@@ -271,6 +280,18 @@ static void *merge_core_dir_configs(pool *a, void *basev, void *newv)
     }
 #endif
 
+    if (new->server_signature != srv_sig_unset) {
+	conf->server_signature = new->server_signature;
+    }
+
+    if (new->add_default_charset != ADD_DEFAULT_CHARSET_UNSET) {
+	conf->add_default_charset = new->add_default_charset;
+    }
+
+    if (new->add_default_charset_name) {
+	conf->add_default_charset_name = new->add_default_charset_name;
+    }
+
     return (void*)conf;
 }
 
@@ -333,7 +354,7 @@ CORE_EXPORT(void) ap_add_per_url_conf(server_rec *s, void *url_config)
     *new_space = url_config;
 }
 
-static void add_file_conf(core_dir_config *conf, void *url_config)
+CORE_EXPORT(void) ap_add_file_conf(core_dir_config *conf, void *url_config)
 {
     void **new_space = (void **)ap_push_array(conf->sec);
     
@@ -407,18 +428,16 @@ void ap_core_reorder_directories(pool *p, server_rec *s)
     int nelts;
     void **elts;
     int i;
+    pool *tmp;
 
-    /* XXX: we are about to waste some ram ... we will build a new array
-     * and we need some scratch space to do it.  The old array and the
-     * scratch space are never freed.
-     */
     sconf = ap_get_module_config(s->module_config, &core_module);
     sec = sconf->sec;
     nelts = sec->nelts;
     elts = (void **)sec->elts;
 
-    /* build our sorting space */
-    sortbin = ap_palloc(p, sec->nelts * sizeof(*sortbin));
+    /* we have to allocate tmp space to do a stable sort */
+    tmp = ap_make_sub_pool(p);
+    sortbin = ap_palloc(tmp, sec->nelts * sizeof(*sortbin));
     for (i = 0; i < nelts; ++i) {
 	sortbin[i].orig_index = i;
 	sortbin[i].elt = elts[i];
@@ -426,15 +445,12 @@ void ap_core_reorder_directories(pool *p, server_rec *s)
 
     qsort(sortbin, nelts, sizeof(*sortbin), reorder_sorter);
 
-    /* and now build a new array */
-    /* XXX: uh I don't see why we can't reuse the old array, what
-     * was I thinking? -djg */
-    sec = ap_make_array(p, nelts, sizeof(void *));
+    /* and now copy back to the original array */
     for (i = 0; i < nelts; ++i) {
-	*(void **)ap_push_array(sec) = sortbin[i].elt;
+      elts[i] = sortbin[i].elt;
     }
 
-    sconf->sec = sec;
+    ap_destroy_pool(tmp);
 }
 
 /*****************************************************************
@@ -667,17 +683,50 @@ API_EXPORT(const char *) ap_get_remote_logname(request_rec *r)
  * name" as supplied by a possible Host: header or full URI.  We never
  * trust the port passed in the client's headers, we always use the
  * port of the actual socket.
+ *
+ * The DNS option to UseCanonicalName causes this routine to do a
+ * reverse lookup on the local IP address of the connectiona and use
+ * that for the ServerName. This makes its value more reliable while
+ * at the same time allowing Demon's magic virtual hosting to work.
+ * The assumption is that DNS lookups are sufficiently quick...
+ * -- fanf 1998-10-03
  */
-API_EXPORT(const char *) ap_get_server_name(const request_rec *r)
+API_EXPORT(const char *) ap_get_server_name(request_rec *r)
 {
+    conn_rec *conn = r->connection;
     core_dir_config *d;
 
     d = (core_dir_config *)ap_get_module_config(r->per_dir_config,
 						&core_module);
-    if (d->use_canonical_name & 1) {
-	return r->server->server_hostname;
+
+    if (d->use_canonical_name == USE_CANONICAL_NAME_OFF) {
+        return r->hostname ? r->hostname : r->server->server_hostname;
     }
-    return r->hostname ? r->hostname : r->server->server_hostname;
+    if (d->use_canonical_name == USE_CANONICAL_NAME_DNS) {
+        if (conn->local_host == NULL) {
+	    struct in_addr *iaddr;
+	    struct hostent *hptr;
+            int old_stat;
+	    old_stat = ap_update_child_status(conn->child_num,
+					      SERVER_BUSY_DNS, r);
+	    iaddr = &(conn->local_addr.sin_addr);
+	    hptr = gethostbyaddr((char *)iaddr, sizeof(struct in_addr),
+				 AF_INET);
+	    if (hptr != NULL) {
+	        conn->local_host = ap_pstrdup(conn->pool,
+					      (void *)hptr->h_name);
+		ap_str_tolower(conn->local_host);
+	    }
+	    else {
+	        conn->local_host = ap_pstrdup(conn->pool,
+					      r->server->server_hostname);
+	    }
+	    (void) ap_update_child_status(conn->child_num, old_stat, r);
+	}
+	return conn->local_host;
+    }
+    /* default */
+    return r->server->server_hostname;
 }
 
 API_EXPORT(unsigned) ap_get_server_port(const request_rec *r)
@@ -688,38 +737,21 @@ API_EXPORT(unsigned) ap_get_server_port(const request_rec *r)
     
     port = r->server->port ? r->server->port : ap_default_port(r);
 
-    if (d->use_canonical_name & 1) {
-	return port;
+    if (d->use_canonical_name == USE_CANONICAL_NAME_OFF
+	|| d->use_canonical_name == USE_CANONICAL_NAME_DNS) {
+        return r->hostname ? ntohs(r->connection->local_addr.sin_port)
+			   : port;
     }
-    return r->hostname ? ntohs(r->connection->local_addr.sin_port)
-		       : port;
+    /* default */
+    return port;
 }
 
 API_EXPORT(char *) ap_construct_url(pool *p, const char *uri,
-				    const request_rec *r)
+				    request_rec *r)
 {
-    unsigned port;
-    const char *host;
-    core_dir_config *d =
-      (core_dir_config *)ap_get_module_config(r->per_dir_config, &core_module);
+    unsigned port = ap_get_server_port(r);
+    const char *host = ap_get_server_name(r);
 
-    if (d->use_canonical_name & 1) {
-	port = r->server->port ? r->server->port : ap_default_port(r);
-	host = r->server->server_hostname;
-    }
-    else {
-        if (r->hostname) {
-            port = ntohs(r->connection->local_addr.sin_port);
-	}
-        else if (r->server->port) {
-            port = r->server->port;
-	}
-        else {
-            port = ap_default_port(r);
-	}
-
-	host = r->hostname ? r->hostname : r->server->server_hostname;
-    }
     if (ap_is_default_port(port, r)) {
 	return ap_pstrcat(p, ap_http_method(r), "://", host, uri, NULL);
     }
@@ -827,7 +859,6 @@ static char* get_interpreter_from_win32_registry(pool *p, const char* ext)
 }
 
 API_EXPORT (file_type_e) ap_get_win32_interpreter(const  request_rec *r, 
-                                                  char*  ext, 
                                                   char** interpreter )
 {
     HANDLE hFile;
@@ -835,50 +866,68 @@ API_EXPORT (file_type_e) ap_get_win32_interpreter(const  request_rec *r,
     BOOLEAN bResult;
     char buffer[1024];
     core_dir_config *d;
-    file_type_e fileType = FileTypeUNKNOWN;
     int i;
+    file_type_e fileType = eFileTypeUNKNOWN;
+    char *ext = NULL;
+    char *exename = NULL;
 
     d = (core_dir_config *)ap_get_module_config(r->per_dir_config, 
                                                 &core_module);
 
-    if (d->script_interpreter_source == INTERPRETER_SOURCE_REGISTRY) {
-        /* 
-         * Check the registry
-         */
+    /* Find the file extension */
+    exename = strrchr(r->filename, '/');
+    if (!exename) {
+        exename = strrchr(r->filename, '\\');
+    }
+    if (!exename) {
+        exename = r->filename;
+    }
+    else {
+        exename++;
+    }
+    ext = strrchr(exename, '.');
+
+    if (ext && (!strcasecmp(ext,".bat") || !strcasecmp(ext,".cmd"))) {
+        return eFileTypeEXE32;
+    }
+
+    /* If the file has an extension and it is not .com and not .exe and
+     * we've been instructed to search the registry, then do it!
+     */
+    if (ext && strcasecmp(ext,".exe") && strcasecmp(ext,".com") &&
+        d->script_interpreter_source == INTERPRETER_SOURCE_REGISTRY) {
+         /* Check the registry */
         *interpreter = get_interpreter_from_win32_registry(r->pool, ext);
         if (*interpreter)
-            return FileTypeSCRIPT;
+            return eFileTypeSCRIPT;
         else {
             ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, r->server,
              "ScriptInterpreterSource config directive set to \"registry\".\n\t"
              "Registry was searched but interpreter not found. Trying the shebang line.");
         }
-    }
+    }        
 
-    /* 
-     * Look for a #! line in the script
-     */
+    /* Need to peek into the file figure out what it really is... */
     hFile = CreateFile(r->filename, GENERIC_READ, FILE_SHARE_READ, NULL,
                        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
-
     if (hFile == INVALID_HANDLE_VALUE) {
-        return FileTypeUNKNOWN;
+        return eFileTypeUNKNOWN;
     }
-
     bResult = ReadFile(hFile, (void*) &buffer, sizeof(buffer) - 1, 
                        &nBytesRead, NULL);
     if (!bResult || (nBytesRead == 0)) {
         ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
                       "ReadFile(%s) failed", r->filename);
         CloseHandle(hFile);
-        return (FileTypeUNKNOWN);
+        return eFileTypeUNKNOWN;
     }
     CloseHandle(hFile);
-    
     buffer[nBytesRead] = '\0';
-    
+
+    /* Script or executable, that is the question... */
     if ((buffer[0] == '#') && (buffer[1] == '!')) {
-        fileType = FileTypeSCRIPT;
+        /* Assuming file is a script since it starts with a shebang */
+        fileType = eFileTypeSCRIPT;
         for (i = 2; i < sizeof(buffer); i++) {
             if ((buffer[i] == '\r')
                 || (buffer[i] == '\n')) {
@@ -886,16 +935,21 @@ API_EXPORT (file_type_e) ap_get_win32_interpreter(const  request_rec *r,
             }
         }
         buffer[i] = '\0';
-        for (i = 2; buffer[i] == ' '; ++i)
+        for (i = 2; buffer[i] == ' ' ; ++i)
             ;
         *interpreter = ap_pstrdup(r->pool, buffer + i ); 
     }
     else {
-        /* Check to see if it's a executable */
-        IMAGE_DOS_HEADER *hdr = (IMAGE_DOS_HEADER*)buffer;
-        if (hdr->e_magic == IMAGE_DOS_SIGNATURE && hdr->e_cblp < 512) {
-            fileType = FileTypeEXE;
+        /* Not a script, is it an executable? */
+        IMAGE_DOS_HEADER *hdr = (IMAGE_DOS_HEADER*)buffer;    
+        if ((nBytesRead >= sizeof(IMAGE_DOS_HEADER)) && (hdr->e_magic == IMAGE_DOS_SIGNATURE)) {
+            if (hdr->e_lfarlc < 0x40)
+                fileType = eFileTypeEXE16;
+            else
+                fileType = eFileTypeEXE32;
         }
+        else
+            fileType = eFileTypeUNKNOWN;
     }
 
     return fileType;
@@ -992,6 +1046,27 @@ static const char *set_gprof_dir(cmd_parms *cmd, void *dummy, char *arg)
 }
 #endif /*GPROF*/
 
+static const char *set_add_default_charset(cmd_parms *cmd, 
+	core_dir_config *d, char *arg)
+{
+    const char *err = ap_check_cmd_context(cmd, NOT_IN_LIMIT);
+    if (err != NULL) {
+        return err;
+    }
+    if (!strcasecmp(arg, "Off")) {
+       d->add_default_charset = ADD_DEFAULT_CHARSET_OFF;
+    }
+    else if (!strcasecmp(arg, "On")) {
+       d->add_default_charset = ADD_DEFAULT_CHARSET_ON;
+       d->add_default_charset_name = DEFAULT_ADD_DEFAULT_CHARSET_NAME;
+    }
+    else {
+       d->add_default_charset = ADD_DEFAULT_CHARSET_ON;
+       d->add_default_charset_name = arg;
+    }
+    return NULL;
+}
+
 static const char *set_document_root(cmd_parms *cmd, void *dummy, char *arg)
 {
     void *sconf = cmd->server->module_config;
@@ -1004,7 +1079,7 @@ static const char *set_document_root(cmd_parms *cmd, void *dummy, char *arg)
     }
 
     arg = ap_os_canonical_filename(cmd->pool, arg);
-    if (!ap_is_directory(arg)) {
+    if (ap_configtestonly && ap_docrootcheck && !ap_is_directory(arg)) {
 	if (cmd->server->is_virtual) {
 	    fprintf(stderr, "Warning: DocumentRoot [%s] does not exist\n",
 		    arg);
@@ -1366,10 +1441,6 @@ static const char *dirsection(cmd_parms *cmd, void *dummy, const char *arg)
     *endp = '\0';
 
     cmd->path = ap_getword_conf(cmd->pool, &arg);
-#ifdef OS2
-    /* Fix OS/2 HPFS filename case problem. */
-    cmd->path = strlwr(cmd->path);
-#endif    
     cmd->override = OR_ALL|ACCESS_CONF;
 
     if (thiscmd->cmd_data) { /* <DirectoryMatch> */
@@ -1543,7 +1614,7 @@ static const char *filesection(cmd_parms *cmd, core_dir_config *c,
     conf->d_is_fnmatch = ap_is_fnmatch(conf->d) != 0;
     conf->r = r;
 
-    add_file_conf(c, new_file_conf);
+    ap_add_file_conf(c, new_file_conf);
 
     if (*arg != '\0') {
 	return ap_pstrcat(cmd->pool, "Multiple ", thiscmd->name,
@@ -2106,15 +2177,25 @@ static const char *set_content_md5(cmd_parms *cmd, core_dir_config *d, int arg)
 }
 
 static const char *set_use_canonical_name(cmd_parms *cmd, core_dir_config *d, 
-					  int arg)
+					  char *arg)
 {
     const char *err = ap_check_cmd_context(cmd, NOT_IN_LIMIT);
-
     if (err != NULL) {
 	return err;
     }
-    
-    d->use_canonical_name = arg != 0;
+
+    if (strcasecmp(arg, "on") == 0) {
+        d->use_canonical_name = USE_CANONICAL_NAME_ON;
+    }
+    else if (strcasecmp(arg, "off") == 0) {
+        d->use_canonical_name = USE_CANONICAL_NAME_OFF;
+    }
+    else if (strcasecmp(arg, "dns") == 0) {
+        d->use_canonical_name = USE_CANONICAL_NAME_DNS;
+    }
+    else {
+        return "parameter must be 'on', 'off', or 'dns'";
+    }
     return NULL;
 }
 
@@ -2122,6 +2203,8 @@ static const char *set_daemons_to_start(cmd_parms *cmd, void *dummy, char *arg)
 {
 #ifdef WIN32
     fprintf(stderr, "WARNING: StartServers has no effect on Win32\n");
+#elif defined(NETWARE)
+    fprintf(stderr, "WARNING: StartServers has no effect on NetWare\n");
 #else
     const char *err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
     if (err != NULL) {
@@ -2341,12 +2424,25 @@ static const char *set_bind_address(cmd_parms *cmd, void *dummy, char *arg)
     return NULL;
 }
 
+#ifdef NETWARE
+static const char *set_threadstacksize(cmd_parms *cmd, void *dummy, char *stacksize)
+{
+    const char *err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
+    if (err != NULL) {
+        return err;
+    }
+    
+    ap_thread_stack_size = atoi(stacksize);    
+    return NULL;
+}
+#endif
+
 static const char *set_listener(cmd_parms *cmd, void *dummy, char *ips)
 {
     listen_rec *new;
     char *ports;
     unsigned short port;
-
+    
     const char *err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
     if (err != NULL) {
         return err;
@@ -2483,7 +2579,8 @@ API_EXPORT(const char *) ap_psignature(const char *prefix, request_rec *r)
 
     conf = (core_dir_config *)ap_get_module_config(r->per_dir_config,
 						   &core_module);
-    if (conf->server_signature == srv_sig_off) {
+    if ((conf->server_signature == srv_sig_off)
+	    || (conf->server_signature == srv_sig_unset)) {
 	return "";
     }
 
@@ -2721,6 +2818,8 @@ static const command_rec core_cmds[] = {
 { "GprofDir", set_gprof_dir, NULL, RSRC_CONF, TAKE1,
   "Directory to plop gmon.out files" },
 #endif
+{ "AddDefaultCharset", set_add_default_charset, NULL, OR_FILEINFO, 
+  TAKE1, "The name of the default charset to add to any Content-Type without one or 'Off' to disable" },
 
 /* Old resource config file commands */
   
@@ -2757,7 +2856,7 @@ static const command_rec core_cmds[] = {
 { "ServerName", set_server_string_slot,
   (void *)XtOffsetOf (server_rec, server_hostname), RSRC_CONF, TAKE1,
   "The hostname of the server" },
-{ "ServerSignature", set_signature_flag, NULL, ACCESS_CONF|RSRC_CONF, TAKE1,
+{ "ServerSignature", set_signature_flag, NULL, OR_ALL, TAKE1,
   "En-/disable server signature (on|off|email)" },
 { "ServerRoot", set_server_root, NULL, RSRC_CONF, TAKE1,
   "Common directory of server-related files (logs, confs, etc.)" },
@@ -2792,9 +2891,8 @@ static const command_rec core_cmds[] = {
 { "ContentDigest", set_content_md5, NULL, OR_OPTIONS,
   FLAG, "whether or not to send a Content-MD5 header with each request" },
 { "UseCanonicalName", set_use_canonical_name, NULL,
-  OR_OPTIONS, FLAG,
-  "Whether or not to always use the canonical ServerName : Port when "
-  "constructing URLs" },
+  RSRC_CONF|ACCESS_CONF, TAKE1,
+  "How to work out the ServerName : Port when constructing URLs" },
 { "StartServers", set_daemons_to_start, NULL, RSRC_CONF, TAKE1,
   "Number of child processes launched at server startup" },
 { "MinSpareServers", set_min_free_servers, NULL, RSRC_CONF, TAKE1,
@@ -2832,6 +2930,10 @@ static const command_rec core_cmds[] = {
    OR_ALL, TAKE12, "soft/hard limits for max number of processes per uid" },
 { "BindAddress", set_bind_address, NULL, RSRC_CONF, TAKE1,
   "'*', a numeric IP address, or the name of a host with a unique IP address"},
+#ifdef NETWARE
+{ "ThreadStackSize", set_threadstacksize, NULL, RSRC_CONF, TAKE1,
+  "Stack size each created thread will use."},
+#endif
 { "Listen", set_listener, NULL, RSRC_CONF, TAKE1,
   "A port number or a numeric IP address and a port number"},
 { "SendBufferSize", set_send_buffer_size, NULL, RSRC_CONF, TAKE1,
@@ -2888,7 +2990,7 @@ static int core_translate(request_rec *r)
     void *sconf = r->server->module_config;
     core_server_config *conf = ap_get_module_config(sconf, &core_module);
   
-    if (r->proxyreq) {
+    if (r->proxyreq != NOT_PROXY) {
         return HTTP_FORBIDDEN;
     }
     if ((r->uri[0] != '/') && strcmp(r->uri, "*")) {
@@ -2928,14 +3030,14 @@ static int core_translate(request_rec *r)
 static int do_nothing(request_rec *r) { return OK; }
 
 #ifdef USE_MMAP_FILES
-struct mmap {
+struct mmap_rec {
     void *mm;
     size_t length;
 };
 
 static void mmap_cleanup(void *mmv)
 {
-    struct mmap *mmd = mmv;
+    struct mmap_rec *mmd = mmv;
 
     if (munmap(mmd->mm, mmd->length) == -1) {
         ap_log_error(APLOG_MARK, APLOG_ERR, NULL,
@@ -2962,6 +3064,9 @@ static int default_handler(request_rec *r)
 #ifdef USE_MMAP_FILES
     caddr_t mm;
 #endif
+#ifdef CHARSET_EBCDIC
+    int convert_flag;
+#endif
 
     /* This handler has no use for a request body (yet), but we still
      * need to read and discard it if the client sent one.
@@ -2985,23 +3090,17 @@ static int default_handler(request_rec *r)
     }
 
     if (r->finfo.st_mode == 0 || (r->path_info && *r->path_info)) {
-	char *emsg;
-
-	emsg = "File does not exist: ";
-	if (r->path_info == NULL) {
-	    emsg = ap_pstrcat(r->pool, emsg, r->filename, NULL);
-	}
-	else {
-	    emsg = ap_pstrcat(r->pool, emsg, r->filename, r->path_info, NULL);
-	}
-	ap_log_rerror(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, r, "%s", emsg);
+	ap_log_rerror(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, r,
+		      "File does not exist: %s",r->path_info ?
+		      ap_pstrcat(r->pool, r->filename, r->path_info, NULL)
+		      : r->filename);
 	return HTTP_NOT_FOUND;
     }
     if (r->method_number != M_GET) {
         return METHOD_NOT_ALLOWED;
     }
 	
-#if defined(OS2) || defined(WIN32)
+#if defined(OS2) || defined(WIN32) || defined(NETWARE)
     /* Need binary mode for OS/2 */
     f = ap_pfopen(r->pool, r->filename, "rb");
 #else
@@ -3023,9 +3122,23 @@ static int default_handler(request_rec *r)
         return errstatus;
     }
 
+#ifdef CHARSET_EBCDIC
+    /* To make serving of "raw ASCII text" files easy (they serve faster
+     * since they don't have to be converted from EBCDIC), a new
+     * "magic" type prefix was invented: text/x-ascii-{plain,html,...}
+     * If we detect one of these content types here, we simply correct
+     * the type to the real text/{plain,html,...} type. Otherwise, we
+     * set a flag that translation is required later on.
+     *
+     * Note: convert_flag is not used in the MMAP path;
+     * ap_checkconv() sets a request_req flag based on content_type
+     */ 
+    convert_flag = ap_checkconv(r);
+#endif
 #ifdef USE_MMAP_FILES
     ap_block_alarms();
     if ((r->finfo.st_size >= MMAP_THRESHOLD)
+	&& (r->finfo.st_size < MMAP_LIMIT)
 	&& (!r->header_only || (d->content_md5 & 1))) {
 	/* we need to protect ourselves in case we die while we've got the
  	 * file mmapped */
@@ -3044,22 +3157,19 @@ static int default_handler(request_rec *r)
 	ap_unblock_alarms();
 #endif
 
+#ifdef CHARSET_EBCDIC
+	if (d->content_md5 & 1) {
+	    ap_table_setn(r->headers_out, "Content-MD5",
+			  ap_md5digest(r->pool, f, convert_flag));
+	}
+#else
 	if (d->content_md5 & 1) {
 	    ap_table_setn(r->headers_out, "Content-MD5",
 			  ap_md5digest(r->pool, f));
 	}
+#endif /* CHARSET_EBCDIC */
 
 	rangestatus = ap_set_byterange(r);
-#ifdef CHARSET_EBCDIC
-	/* To make serving of "raw ASCII text" files easy (they serve faster 
-	 * since they don't have to be converted from EBCDIC), a new
-	 * "magic" type prefix was invented: text/x-ascii-{plain,html,...}
-	 * If we detect one of these content types here, we simply correct
-	 * the type to the real text/{plain,html,...} type. Otherwise, we
-	 * set a flag that translation is required later on.
-	 */
-        ap_checkconv(r);
-#endif /*CHARSET_EBCDIC*/
 
 	ap_send_http_header(r);
 	
@@ -3089,7 +3199,7 @@ static int default_handler(request_rec *r)
 #ifdef USE_MMAP_FILES
     }
     else {
-	struct mmap *mmd;
+	struct mmap_rec *mmd;
 
 	mmd = ap_palloc(r->pool, sizeof(*mmd));
 	mmd->mm = mm;
@@ -3101,7 +3211,7 @@ static int default_handler(request_rec *r)
 	    AP_MD5_CTX context;
 	    
 	    ap_MD5Init(&context);
-	    ap_MD5Update(&context, (void *)mm, r->finfo.st_size);
+	    ap_MD5Update(&context, (void *)mm, (unsigned int)r->finfo.st_size);
 	    ap_table_setn(r->headers_out, "Content-MD5",
 			  ap_md5contextTo64(r->pool, &context));
 	}
