@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1999 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -284,7 +284,7 @@ static char *check_code(pool *p, const char *code, char **real_code)
     while (word[0]) {
         /* <num>
          */
-        if (isdigit(word[0])) {
+        if (ap_isdigit(word[0])) {
             num = atoi(word);
         }
         else {
@@ -407,18 +407,15 @@ static int add_expires(request_rec *r)
     time_t expires;
     char age[20];
 
-    if (is_HTTP_ERROR(r->status))       /* Don't add Expires headers to errors */
+    if (ap_is_HTTP_ERROR(r->status))       /* Don't add Expires headers to errors */
         return DECLINED;
 
     if (r->main != NULL)        /* Say no to subrequests */
         return DECLINED;
 
-    if (r->finfo.st_mode == 0)  /* no file ? shame. */
-        return DECLINED;
-
     conf = (expires_dir_config *) ap_get_module_config(r->per_dir_config, &expires_module);
     if (conf == NULL) {
-        ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
+        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
                     "internal error: %s", r->filename);
         return SERVER_ERROR;
     };
@@ -440,7 +437,8 @@ static int add_expires(request_rec *r)
     if (r->content_type == NULL)
         code = NULL;
     else
-        code = (char *) ap_table_get(conf->expiresbytype, r->content_type);
+        code = (char *) ap_table_get(conf->expiresbytype, 
+		ap_field_noparam(r->pool, r->content_type));
 
     if (code == NULL) {
         /* no expires defined for that type, is there a default? */
@@ -454,6 +452,12 @@ static int add_expires(request_rec *r)
 
     switch (code[0]) {
     case 'M':
+	if (r->finfo.st_mode == 0) { 
+	    /* file doesn't exist on disk, so we can't do anything based on
+	     * modification time.  Note that this does _not_ log an error.
+	     */
+	    return DECLINED;
+	}
         base = r->finfo.st_mtime;
         additional = atoi(&code[1]);
         break;
@@ -468,7 +472,7 @@ static int add_expires(request_rec *r)
         /* expecting the add_* routines to be case-hardened this 
          * is just a reminder that module is beta
          */
-        ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
+        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
                     "internal error: bad expires code: %s", r->filename);
         return SERVER_ERROR;
     };
@@ -505,3 +509,11 @@ module MODULE_VAR_EXPORT expires_module =
     NULL,                       /* child_exit */
     NULL                        /* post read-request */
 };
+
+
+#ifdef NETWARE
+int main(int argc, char *argv[]) 
+{
+    ExitThread(TSR_THREAD, 0);
+}
+#endif
